#!/usr/bin/env python3
#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Use a calibrated USRP as a power meter
"""

import sys
import signal
import argparse
import uhd

def parse_args():
    """Parse the command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--args", default="",
                        help="USRP Device Args")
    parser.add_argument("-f", "--freq", type=float, required=True,
                        help="Center Frequency")
    parser.add_argument("-o", "--lo-offset", type=float, default=0.0,
                        help="Optional LO offset")
    parser.add_argument("-c", "--channel", type=int, default=0,
                        help="USRP RX Channel Index")
    parser.add_argument("-t", "--antenna",
                        help="USRP RX Antenna")
    parser.add_argument("-r", "--rate", default=1e6, type=float,
                        help="Sampling Rate")
    parser.add_argument("-b", "--bandwidth", type=float,
                        help="Analog filter bandwidth (if supported)")
    parser.add_argument("-l", "--ref-level", type=float, default=-15,
                        help="RX reference power level. "
                             "This should be higher than the expected power.")
    parser.add_argument("-n", "--samps-per-est", type=float, default=1e6,
                        help="Samples per estimate.")
    parser.add_argument("--mode", choices=['one-shot', 'continuous'], default='one-shot',
                        help="Measure once, or keep measuring until Ctrl-C is pressed.")
    return parser.parse_args()

def get_streamer(usrp, chan):
    """
    Return an RX streamer with fc32 output
    """
    stream_args = uhd.usrp.StreamArgs('fc32', 'sc16')
    stream_args.channels = [chan]
    return usrp.get_rx_stream(stream_args)

def setup_device(usrp, args):
    """
    Apply the settings from args to the device
    """
    chan = args.channel
    if chan > usrp.get_rx_num_channels():
        print("ERROR: Invalid channel selected: {} (only {} channels available!)"
              .format(chan, usrp.get_rx_num_channels()))
        raise RuntimeError("Invalid channel selected")
    print("Using channel: {}".format(chan))

    if args.antenna:
        print("Setting RX antenna to `{}'...".format(args.antenna), end='')
        usrp.set_rx_antenna(args.antenna, chan)
        print("OK")

    if not usrp.has_rx_power_reference(chan):
        antenna = usrp.get_rx_antenna()
        print("ERROR: This device is not calibrated for RX at RF%d-%s!" % (chan, antenna))
        raise RuntimeError("Device not calibrated for selected antenna")

    print("Requesting RX rate of {} Msps...".format(args.rate / 1e6), end='')
    usrp.set_rx_rate(args.rate, chan)
    if abs(usrp.get_rx_rate(chan) - args.rate) > 1.0:
        print("ALMOST. Actual rate: {} Msps"
              .format(usrp.get_rx_rate(chan) / 1e6))
    else:
        print("OK")

    print("Requesting RX frequency of {} MHz...".format(args.freq / 1e6), end='')
    tr = uhd.types.TuneRequest(args.freq, args.lo_offset)
    usrp.set_rx_freq(tr, chan)
    if abs(usrp.get_rx_freq(chan) - args.freq) > 1.0:
        print("ALMOST. Actual frequency: {} MHz".format(usrp.get_rx_freq(chan) / 1e6))
    else:
        print("OK")

    print("Requesting RX power reference level of {:.2f} dBm..."
          .format(args.ref_level), end='')
    usrp.set_rx_power_reference(args.ref_level, chan)
    ref_level = usrp.get_rx_power_reference(chan)
    if abs(ref_level - args.ref_level) > 1.0:
        print("ALMOST. Actual ref level: {:.2f} dBm".format(ref_level))
    else:
        print("OK")

    if args.bandwidth:
        print("Requesting analog RX bandwidth of {} Msps..."
              .format(args.bandwidth), end='')
        usrp.set_rx_bandwidth(args.bandwidth, chan)
        if abs(usrp.get_rx_bandwidth(chan) - args.bandwidth) > 1.0:
            print("ALMOST. Actual bandwidth: {} MHz"
                  .format(usrp.get_rx_bandwidth(chan) / 1e6))
        else:
            print("OK")
    return (chan, ref_level)

RUN = True

def main():
    """
    gogogo
    """
    args = parse_args()
    usrp = uhd.usrp.MultiUSRP(args.args)
    (chan, ref_level) = setup_device(usrp, args)
    streamer = get_streamer(usrp, chan)
    if args.mode == 'continuous':
        def handle_sigint(_sig, _frame):
            print("Caught Ctrl-C, exiting...")
            global RUN
            RUN = False
        signal.signal(signal.SIGINT, handle_sigint)
    while RUN:
        try:
            power_dbfs = uhd.dsp.signals.get_usrp_power(
                streamer, num_samps=int(args.samps_per_est))
        except RuntimeError:
            # This is a hack b/c the signal handler is not gracefully handling
            # SIGINT
            break
        power_dbm = power_dbfs + ref_level
        print("Received power: {:+6.2f} dBm".format(power_dbm))
        if args.mode == 'one-shot':
            break
    return True

if __name__ == "__main__":
    sys.exit(not main())



#!/usr/bin/env python3
#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Utility to run power calibrations with USRPs
"""

import sys
import time
import math
import pickle
import argparse
import uhd

def parse_args():
    """ Parse args and return args object """
    parser = argparse.ArgumentParser(
        description="Run power level calibration for supported USRPs",
    )
    parser.add_argument(
        '--args', default="",
        help="USRP Device Args",
    )
    parser.add_argument(
        '-d', '--dir', default='tx',
        help="Direction: Must be either rx or tx. From the perspective of the "
             "device: rx == receive power levels, tx == transmit power levels.",
        choices=['tx', 'rx'])
    parser.add_argument(
        '--start',
        help='Start Frequency. Defaults to lowest available frequency on device. '
             'Note that this is only a hint for the device object, which can choose '
             'to override this value.', type=float)
    parser.add_argument(
        '--stop',
        help='Stop Frequency. Defaults to highest available frequency on device. '
             'Note that this is only a hint for the device object, which can choose '
             'to override this value.', type=float)
    parser.add_argument(
        '--step',
        help='Frequency Step. Defaults are device-specific. '
             'Note that this is only a hint for the device object, which can choose '
             'to override this value. Devices can also measure at non-regular '
             'frequencies, e.g., to more accurately cover differences between bands.',
        type=float)
    parser.add_argument(
        '--gain-step', type=float, default=1,
        help='Gain Step (dB). Defaults to 1 dB. '
             'Note that this is only a hint for the device object, which can choose '
             'to override this value. Devices can also measure at non-regular '
             'gain intervals.')
    parser.add_argument(
        '--lo-offset', type=float,
        help='LO Offset. This gets applied to every tune request. Note that for '
             'TX measurements, there is also an offset applied by --tone-freq. '
             'The default value is device-dependent.')
    parser.add_argument(
        '--amplitude', type=float, default=1./math.sqrt(2),
        help='Amplitude of the tone that is generated for tx measurements. '
             'Default is 1/sqrt(2), or -3 dBFS.')
    parser.add_argument(
        '--attenuation', type=float, default=0.0,
        help='Amount of attenuation between measurement device and DUT. This will '
             'be accounted for by simple addition, it is treated like a measurement error. '
             'The argument is generally positive, e.g. 30 means 30 dB of attenuation.')
    parser.add_argument(
        '--tone-freq', type=float, default=1e6,
        help='Frequency of the tone that is generated for Tx measurements. This '
             'has the same effect as setting an LO offset, except in software.')
    parser.add_argument(
        '--antenna', default="*",
        help="Select antenna port. A value of '*' means that the calibration "
             "will be repeated on all appropriate antenna ports.")
    parser.add_argument(
        '--channels', default="*",
        help="Select channel. A value of '*' means that the calibration "
             "will be repeated on all appropriate channels.")
    parser.add_argument(
        '--meas-dev', default='manual',
        help='Type of measurement device that is used')
    parser.add_argument(
        '-o', '--meas-option', default=[], action='append',
        help='Options that are passed to the measurement device')
    parser.add_argument(
        '--switch', default='manual',
        help='Type of switch to be used to connect antennas')
    parser.add_argument(
        '--switch-option', default=[], action='append',
        help='Options that are passed to the switch')
    parser.add_argument(
        '-r', '--rate', type=float,
        help='Sampling rate at which the calibration is performed')
    parser.add_argument(
        '--store', metavar='filename.pickle',
        help='If provided, will store intermediate cal data. This can be analyzed '
        'separately, or loaded into the tool with --load.')
    parser.add_argument(
        '--load', metavar='filename.pickle',
        help='If provided, will load intermediate cal data instead of running a '
        'measurement.')
    return parser.parse_args()


def sanitize_args(usrp, args, default_rate):
    """
    Check the args against the USRP object.
    """
    assert usrp.get_num_mboards() == 1, \
        "Power calibration tools are designed for a single motherboard!"
    available_chans = getattr(usrp, 'get_{}_num_channels'.format(args.dir))()
    if args.channels == '*':
        # * means all channels
        channels = list(range(available_chans))
    else:
        try:
            channels = [int(c) for c in args.channels.split(',')]
        except ValueError:
            raise ValueError("Invalid channel list: {}".format(args.channels))
        for chan in channels:
            assert chan in range(available_chans), \
                "ERROR: Invalid channel: {}. Should be in 0..{}.".format(
                    chan, available_chans)
    print("=== Calibrating for channels:", ", ".join([str(x) for x in channels]))
    available_ants = getattr(usrp, 'get_{}_antennas'.format(args.dir))()
    if args.antenna == '*':
        invalid_antennas = ('CAL', 'LOCAL', 'CAL_LOOPBACK', 'TERMINATION')
        antennas = [x for x in available_ants if x not in invalid_antennas]
    else:
        try:
            antennas = args.antenna.split(',')
        except ValueError:
            raise ValueError("Invalid antenna list: {}".format(args.antenna))
        for ant in antennas:
            assert ant in available_ants, \
                "Invalid antenna: {}. Should be in {}.".format(
                    ant, available_ants)
    print("=== Calibrating for antennas:", ", ".join([str(x) for x in antennas]))
    rate = args.rate or default_rate
    getattr(usrp, 'set_{}_rate'.format(args.dir))(rate)
    actual_rate = getattr(usrp, 'get_{}_rate'.format(args.dir))()
    print("=== Requested sampling rate: {} Msps, actual rate: {} Msps"
          .format(rate/1e6, actual_rate/1e6))
    if args.dir == 'tx' and abs(args.tone_freq) > actual_rate:
        raise ValueError(
            "The TX tone frequency offset of {} kHz is greater than the sampling rate."
            .format(args.tone_freq / 1e3))
    return channels, antennas, actual_rate


def init_results(pickle_file):
    """
    Initialize results from pickle file, or empty dict
    """
    if pickle_file is None:
        return {}
    with open(pickle_file, 'rb') as results_file:
        return pickle.load(results_file)

class CalRunner:
    """
    Executor of the calibration routines.
    """
    def __init__(self, usrp, usrp_cal, meas_dev, args):
        self.usrp = usrp
        self.usrp_cal = usrp_cal
        self.meas_dev = meas_dev
        self.dir = args.dir
        self.tone_offset = args.tone_freq if args.dir == 'tx' else 0.0
        self.lo_offset = args.lo_offset if args.lo_offset else usrp_cal.lo_offset
        if self.lo_offset:
            print("=== Using USRP LO offset: {:.2f} MHz"
                  .format(self.lo_offset / 1e6))

    def run(self, chan, freq):
        """
        Run all cal steps for a single frequency
        """
        print("=== Running calibration at frequency {:.3f} MHz...".format(freq / 1e6))
        tune_req = uhd.types.TuneRequest(freq, self.lo_offset)
        getattr(self.usrp, 'set_{}_freq'.format(self.dir))(tune_req, chan)
        time.sleep(self.usrp_cal.tune_settling_time)
        actual_freq = getattr(self.usrp, 'get_{}_freq'.format(self.dir))(chan)
        if abs(actual_freq - freq) > 1.0:
            print("WARNING: Frequency was coerced from {:.2f} MHz to {:.2f} MHz!"
                  .format(freq / 1e6, actual_freq / 1e6))
        self.meas_dev.set_frequency(actual_freq + self.tone_offset)
        getattr(self.usrp_cal, 'run_{}_cal'.format(self.dir))(freq)

def main():
    """Go, go, go!"""
    args = parse_args()
    print("=== Detecting USRP...")
    usrp = uhd.usrp.MultiUSRP(args.args)
    print("=== Measurement direction:", args.dir)
    print("=== Initializing measurement device...")
    meas_dev = uhd.usrp.cal.get_meas_device(args.dir, args.meas_dev, args.meas_option)
    meas_dev.power_offset = args.attenuation
    # If we're transmitting, then we need to factor in the "attenuation" from us
    # not transmitting at full scale
    if args.dir == 'tx':
        meas_dev.power_offset -= 20 * math.log10(args.amplitude)
    print("=== Initializing port connector...")
    switch = uhd.usrp.cal.get_switch(args.dir, args.switch, args.switch_option)
    print("=== Initializing USRP calibration object...")
    usrp_cal = uhd.usrp.cal.get_usrp_calibrator(
        usrp, meas_dev, args.dir,
        gain_step=args.gain_step,
    )
    channels, antennas, rate = sanitize_args(usrp, args, usrp_cal.default_rate)
    results = init_results(args.load)
    usrp_cal.init(
        rate=rate,
        tone_freq=args.tone_freq,
        amplitude=args.amplitude,
    )
    print("=== Launching calibration...")
    cal_runner = CalRunner(usrp, usrp_cal, meas_dev, args)
    for chan in channels:
        if chan not in results:
            results[chan] = {}
        for ant in antennas:
            if ant in results[chan]:
                print("=== Using pickled data for channel {}, antenna {}."
                      .format(chan, ant))
                continue
            print("=== Running calibration for channel {}, antenna {}."
                  .format(chan, ant))
            # Set up all the objects
            getattr(usrp, 'set_{}_antenna'.format(args.dir))(ant, chan)
            switch.connect(chan, ant)
            usrp_cal.update_port(chan, ant)
            freqs = usrp_cal.init_frequencies(args.start, args.stop, args.step)
            usrp_cal.start() # This will activate siggen
            # Now calibrate
            for freq in freqs:
                try:
                    cal_runner.run(chan, freq)
                except RuntimeError as ex:
                    print("ERROR: Stopping calibration due to exception: {}"
                          .format(str(ex)))
                    usrp_cal.stop()
                    return 1
            # Store results for pickling and shut down for next antenna port
            results[chan][ant] = usrp_cal.results
            usrp_cal.stop() # This will deactivate siggen and store the data
    if args.store:
        print("=== Storing pickled calibration data to {}...".format(args.store))
        with open(args.store, 'wb') as results_file:
            pickle.dump(results, results_file)
    return 0

if __name__ == "__main__":
    try:
        sys.exit(main())
    except (RuntimeError, ValueError) as ex:
        print("ERROR:", str(ex))
        sys.exit(1)



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Measurement Device Class for UHD Power Calibration
"""

import sys
import time
import inspect
import importlib
import numpy
import uhd
from .tone_gen import ToneGenerator
from .visa import get_visa_device
from .ni_rf_instr import RFSADevice, RFSGDevice

###############################################################################
# Base Classes
###############################################################################
class PowerMeterBase:
    """
    Base class for measuring output power (Tx) of the USRP. That means the
    measurement device is receiving and the USRP (the DUT) is transmitting.
    """
    def __init__(self, options):
        self._options = options
        self.power_offset = 0

    def set_frequency(self, freq):
        """
        Set the frequency of the measurement device.
        """
        raise NotImplementedError()

    def get_power(self):
        """
        Return the current measured power in dBm.
        """
        return self._get_power() + self.power_offset

    def _get_power(self):
        """
        Return the current measured power in dBm.
        """
        raise NotImplementedError()

class SignalGeneratorBase:
    """
    Base class for measuring input power (Rx) of the USRP. That means the
    measurement device is transmitting and the USRP (the DUT) is receiving.
    """
    def __init__(self, options):
        self._options = options
        self.power_offset = 0
        # Make sure to set this before doing RX cal
        self.max_output_power = None

    def enable(self, enable=True):
        """
        Turn on the power generator. By default, it should be off, and only
        produce a signal when this was called with an argument value of 'True'.
        """
        raise NotImplementedError()

    def set_power(self, power_dbm):
        """
        Set the input power of the DUT. This will factor in the power offset,
        and set the measurement device to produce the power that will cause the
        DUT to receive power_dbm.

        This will coerce to the next possible power available and return the
        coerced value.
        """
        assert self.max_output_power is not None
        if power_dbm > self.max_output_power:
            print("[SigGen] WARNING! Trying to set power beyond safe levels. "
                  "Capping output power at {} dBm.".format(self.max_output_power))
            power_dbm = self.max_output_power
        return self._set_power(power_dbm + self.power_offset) - self.power_offset

    def get_power(self):
        """
        Return the input power of the DUT.  This will factor in the power offset,
        and will return the power level in dBm that is going into the DUT.
        Use this with set_power(), as not all power levels can be reached.
        """
        return self._get_power() - self.power_offset

    def set_frequency(self, freq):
        """
        Set the center frequency of the generated signal.
        """
        raise NotImplementedError()

    def _set_power(self, power_dbm):
        """
        Set the output power of the device in dBm.
        """
        raise NotImplementedError()

    def _get_power(self):
        """
        Return the output power of the measurement device.
        """
        raise NotImplementedError()

###############################################################################
# Manual Measurement: For masochists, or for small sample sets
###############################################################################
class ManualPowerMeter(PowerMeterBase):
    """
    Manual measurement: The script does nothing, it just asks the user to
    manually make changes and return values
    """
    key = 'manual'

    def set_frequency(self, freq):
        """
        Ask user to set frequency
        """
        input("[TX] Set your power meter to following frequency: "
              "{:.3f} MHz, then hit Enter.".format(freq/1e6))

    def _get_power(self):
        """
        Ask user for the power
        """
        num_tries = 5
        for _ in range(num_tries):
            try:
                return float(input("[TX] Please enter the measured power in dBm: ")) \
                       + self.power_offset
            except ValueError:
                continue
        raise ValueError("Invalid power value entered.")

class ManualPowerGenerator(SignalGeneratorBase):
    """
    Manual measurement: The script does nothing, it just asks the user to
    manually make changes and return values
    """
    key = 'manual'
    num_tries = 5

    def enable(self, enable=True):
        """
        Ask the user to turn the device on or off
        """
        input("[RX] Please {} your signal generator and hit Enter."
              .format("enable" if enable else "disable"))

    def _set_power(self, power_dbm):
        """
        Ask for a power, or the closest, and return that
        """
        new_power = input(
            "[RX] Set your signal generator to following output power: "
            "{:.2f} dBm, then hit Enter, or enter the closest available power: "
            .format(power_dbm))
        if not new_power:
            return power_dbm
        for _ in range(self.num_tries):
            try:
                return float(new_power)
            except ValueError:
                new_power = input(
                    "[RX] Set your signal generator to following output power: "
                    "{:.2f} dBm, then hit Enter, or enter the closest available power: "
                    .format(power_dbm))
                if not new_power:
                    return power_dbm
        raise ValueError("Invalid power value entered.")

    def _get_power(self):
        """
        Ask user for current power
        """
        for _ in range(self.num_tries):
            try:
                return float(input(
                    "[RX] Please enter the output power in dBm of your "
                    "signal generator: "))
            except ValueError:
                continue
        raise ValueError("Invalid power value entered.")

    # pylint: disable=no-self-use
    def set_frequency(self, freq):
        """
        Set the center frequency of the generated signal.
        """
        input("[RX] Set your signal generator to following frequency: {:.3f} MHz, then hit Enter."
              .format(freq/1e6))
    # pylint: enable=no-self-use

##############################################################################
# RFSA: Run through a NI-RFSA device, using RFmx library
###############################################################################
class RfsaPowerMeter(PowerMeterBase):
    """
    Power meter using RFmx TXP measurement on NI-RFSA devices.
    """
    key = 'rfsa'

    def __init__(self, options):
        super().__init__(options)
        self.device = RFSADevice(options)

    def set_frequency(self, freq):
        """
        Set the frequency of the measurement device.
        """
        self.device.set_frequency(freq)

    def _get_power(self):
        """
        Return the current measured power in dBm.
        """
        return self.device.get_power_dbm()

##############################################################################
# VISA: Run through a VISA device, using SCPI commands
###############################################################################
class VisaPowerMeter(PowerMeterBase):
    """
    VISA based Tx measurement device
    """
    DEFAULT_VISA_LIB = '@py' # pyvisa-py
    DEFAULT_VISA_QUERY = "?*::INSTR"

    key = 'visa'

    def __init__(self, options):
        super().__init__(options)
        # pylint: disable=import-outside-toplevel
        # We disable this warning because having pyvisa installed is not a
        # requirement, so we want to load it as late as possible, and only when
        # needed.
        import pyvisa
        # pylint: enable=import-outside-toplevel
        visa_lib = options.get('visa_lib', self.DEFAULT_VISA_LIB)
        visa_query = options.get('visa_query', self.DEFAULT_VISA_QUERY)
        self._rm = pyvisa.ResourceManager(visa_lib)
        resources = self._rm.list_resources(visa_query)
        if len(resources) > 1:
            print("Found VISA devices:")
            for resource in resources:
                print("*" + resource)
            raise RuntimeError(
                "Found more than one measurement device. Please limit the query!")
        if len(resources) == 0:
            raise RuntimeError("No measurement device found!")
        self._res = self._rm.open_resource(resources[0])
        self.visa = get_visa_device(self._res, resources[0], options)
        self.visa.init_power_meter()

    def set_frequency(self, freq):
        """
        Set frequency
        """
        self.visa.set_frequency(freq)

    def _get_power(self):
        """
        Get power
        """
        return self.visa.get_power_dbm()

###############################################################################
# USRP: Use a pre-calibrated USRP as a measurement device
###############################################################################
class USRPPowerGenerator(SignalGeneratorBase):
    """
    The power generator is actually a USRP. This only works if the USRP that is
    used for power/signal generation has been previously calbrated itself.
    """
    key = 'usrp'

    def __init__(self, options):
        super().__init__(options)
        usrp_args = options.get('args')
        if not usrp_args:
            raise RuntimeError(
                "Must specify args for USRP measurement device!")
        self._usrp = uhd.usrp.MultiUSRP(usrp_args)
        self._rate = float(options.get('rate', 5e6))
        self._lo_offset = float(options.get('lo_offset', 0))
        self._chan = int(options.get('chan', 0))
        self._amplitude = float(options.get('ampl', 1/numpy.sqrt(2)))
        self._pwr_dbfs = 20 * numpy.log10(self._amplitude)
        self._tone_freq = 0
        stream_args = uhd.usrp.StreamArgs('fc32', 'sc16')
        stream_args.channels = [self._chan]
        self._streamer = self._usrp.get_tx_stream(stream_args)
        print("==== Creating USRP tone generator. Power offset:", self.power_offset)
        self._tone_gen = ToneGenerator(self._rate, self._tone_freq, self._amplitude)
        self._tone_gen.set_streamer(self._streamer)

    def enable(self, enable=True):
        """
        Turn the tone generator on or off.
        """
        if enable:
            print("[SigGen] Starting tone generator.")
            self._tone_gen.start()
        else:
            print("[SigGen] Stopping tone generator.")
            self._tone_gen.stop()
        time.sleep(0.1) # Give it some time to spin down

    def set_frequency(self, freq):
        """
        Set the center frequency of the generated signal.
        """
        print("[SigGen] Channel {}: Tuning signal to {:.3f} MHz."
              .format(self._chan, freq/1e6))
        tune_req = uhd.types.TuneRequest(freq, self._lo_offset)
        self._usrp.set_tx_freq(tune_req, self._chan)

    def _set_power(self, power_dbm):
        """
        Set the output power of the device in dBm.
        """
        self._usrp.set_tx_power_reference(power_dbm - self._pwr_dbfs, self._chan)
        return self._get_power()

    def _get_power(self):
        """
        Return the output power of the measurement device.
        """
        return self._usrp.get_tx_power_reference(self._chan) + self._pwr_dbfs

###############################################################################
# RFSG: NI signal generator family
###############################################################################
class RFSGPowerGenerator(SignalGeneratorBase):
    """
    Power Generator using NI-RFSG devices.
    """
    key = 'rfsg'

    def __init__(self, options):
        super().__init__(options)
        self.device = RFSGDevice(options)

    def enable(self, enable=True):
        """
        Turn tone generation on and off
        """
        self.device.enable(enable)

    def set_frequency(self, freq):
        """
        Set the center frequency of the generated signal.
        """
        self.device.set_frequency(freq)

    def _set_power(self, power_dbm):
        """
        Set the output power of the device in dBm.
        """
        return self.device.set_power(power_dbm)

    def _get_power(self):
        """
        Get the output power of the device in dBm.
        """
        return self.device.get_power()

###############################################################################
# The dispatch function
###############################################################################
def get_meas_device(direction, dev_key, options):
    """
    Return the measurement device object
    """
    assert direction in ('tx', 'rx')
    base_class = SignalGeneratorBase if direction == 'rx' else PowerMeterBase
    opt_dict = {
        k[0]: k[1] if len(k) > 1 else None for k in [x.split("=", 1) for x in options]
    }
    members = inspect.getmembers(sys.modules[__name__])
    if 'import' in opt_dict:
        try:
            print("Loading external module: {}".format(opt_dict.get('import')))
            external_module = importlib.import_module(opt_dict.get('import'))
            members += inspect.getmembers(external_module)
        except (ModuleNotFoundError, ImportError):
            print("WARNING: Could not import module '{}'"
                  .format(opt_dict.get('import')))
    for _, obj in members:
        try:
            if issubclass(obj, base_class) and dev_key == getattr(obj, 'key', ''):
                return obj(opt_dict)
        except TypeError:
            continue
    raise RuntimeError("No {} found for key: {}".format(
        "signal generator" if direction == "rx" else "power meter", dev_key))



#
# Copyright 2017-2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
""" @package types
Python UHD module containing types to be used with a MultiUSRP object
"""

from . import libpyuhd as lib


StreamMode = lib.types.stream_mode
StreamCMD = lib.types.stream_cmd
DeviceAddr = lib.types.device_addr
TimeSpec = lib.types.time_spec
SPIEdge = lib.types.spi_edge
SPIConfig = lib.types.spi_config
RXMetadataErrorCode = lib.types.rx_metadata_error_code
Range = lib.types.range
MetaRange = lib.types.meta_range_t
GainRange = lib.types.meta_range_t
FreqRange = lib.types.meta_range_t
RXMetadata = lib.types.rx_metadata
TXMetadata = lib.types.tx_metadata
TXAsyncMetadata = lib.types.async_metadata
TXMetadataEventCode = lib.types.tx_metadata_event_code
DataType = lib.types.data_type
SensorValue = lib.types.sensor_value
TuneRequestPolicy = lib.types.tune_request_policy
TuneRequest = lib.types.tune_request
TuneResult = lib.types.tune_result



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Switch Device Classes for UHD Power Calibration
"""

import sys
import inspect
import importlib
from .ni_rf_instr import get_modinst_device

# pylint: disable=too-few-public-methods
class SwitchBase:
    """
    Base class to connect ports of measurement equipment and USRP.
    """
    key = ""

    def connect(self, chan, antenna):
        """
        Connect a port of the USRP (DUT) denoted by chan and antenna to the
        measurement device.
        :param chan: channel to connect
        :param antenna: antenna to connect
        """
        raise NotImplementedError()


class ManualSwitch(SwitchBase):
    """
    ManualSwitch is the fallback implementation if no other switch can be
    found. It asks the user to change cable setup and halts the calibration
    until the user confirms the configuration. If `mode=auto` is given in
    options connect call assumes there is no need to pause for connecting
    measurement device with DUT (e.g. only one path is measured).
    """
    def __init__(self, direction, options=None):
        self.direction = direction
        self.mode = options.get('mode', '')

    def connect(self, chan, antenna):
        """
        Connect a port of the USRP (DUT) denoted by chan and antenna to the
        measurement device. In this manual mode the user is responsible to
        ensure correct wiring. The script waits until the user confirms the
        setup.
        :param chan: channel to connect
        :param antenna: antenna to connect
        """
        if self.mode == 'auto':
            return # no need to wait for manual connection
        dev_type = "signal generator" if self.direction == 'rx' else "power meter"
        input(f"[{self.direction}] Connect your {dev_type} to device channel {chan}, "
              f"antenna {antenna}. Then, hit Enter.")

class NISwitch(SwitchBase):
    """
    Use NI switch devices to automatically connect measurement devices with
    DUT.
    """
    key = "niswitch"

    def __init__(self, options):
        # connections stores the connected ports for each chan/antenna
        # combination. During connect call connection is checked for an
        # appropriate key. If there is no such key the next channel of the
        # current port switch is used.
        # To get a working setup
        self.connections = {}

        device = get_modinst_device("NI-SWITCH", options.get("name", ""))
        # pylint: disable=import-outside-toplevel
        # disable import warning. We do not want to make niswitch a mandatory
        # package for users who do not use this class
        import niswitch
        self.session = niswitch.Session(device.device_name)
        self.port = options.get("port", "comA")

    def connect(self, chan, antenna):
        """
        Connect a port of the USRP (DUT) denoted by chan and antenna to the
        measurement device.
        The NI switch connects the channel of the switch (named "chXY") to the
        port of the switch ("comX") which was given during initialization.
        The connection is made consecutively, meaning the first requested
        configuration of channel and antenna is connected to chX1, the second
        to chX2 and so on. X is the identifier of the port given in
        initialization. The user has to make sure that the SMA port of the DUT
        are cabled in the right order. The calibration loops over all
        (configured) channels and for each channel over the (configured)
        antennas. The order of channels and antennas can be changed via script
        parameter. The loop order is fixed.
        :param chan: channel to connect
        :param antenna: antenna to connect
        """
        key = (chan, antenna)
        if key not in self.connections:
            # connection not known yet, generate next connection pair
            # first item is port used by this instance
            # second item is next channel, channels are named chXY where
            # X is the letter of the current port (derived from port name) and
            # Y is a number starting at 1
            switch_channel = "ch%d%s" % (len(self.connections) + 1,
                                         self.port[-1:])
            self.connections[key] = (self.port, switch_channel)

        connection = self.connections[key]
        print("Connecting %s-%s at switch to measure %d-%s of DUT" %
              (connection[0], connection[1], chan, antenna))
        self.session.disconnect_all()
        self.session.connect(connection[0], connection[1])

###############################################################################
# The dispatch function
###############################################################################
def get_switch(direction, dev_key, options):
    """
    Return the measurement device object
    """
    opt_dict = {
        k[0]: k[1] if len(k) > 1 else None for k in [x.split("=", 1) for x in options]
    }
    members = inspect.getmembers(sys.modules[__name__])
    if 'import' in opt_dict:
        try:
            print("Loading external module: {}".format(opt_dict.get('import')))
            external_module = importlib.import_module(opt_dict.get('import'))
            members += inspect.getmembers(external_module)
        except (ModuleNotFoundError, ImportError):
            print("WARNING: Could not import module '{}'"
                  .format(opt_dict.get('import')))
    for _, obj in members:
        try:
            if issubclass(obj, SwitchBase) and dev_key == getattr(obj, 'key', ''):
                return obj(opt_dict)
        except TypeError:
            continue
    return ManualSwitch(direction, opt_dict)



#!/usr/bin/env python3
#
# Copyright 2018 Ettus Research, a National Instruments Company
# Copyright 2019 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Benchmark rate using Python API
"""

import argparse
from datetime import datetime, timedelta
import sys
import time
import threading
import logging
import numpy as np
import uhd


CLOCK_TIMEOUT = 1000  # 1000mS timeout for external clock locking
INIT_DELAY = 0.05  # 50mS initial delay before transmit

def parse_args():
    """Parse the command line arguments"""
    description = """UHD Benchmark Rate (Python API)

        Utility to stress test a USRP device.

        Specify --rx_rate for a receive-only test.
        Specify --tx_rate for a transmit-only test.
        Specify both options for a full-duplex test.
        """
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description=description)
    parser.add_argument("-a", "--args", default="", type=str, help="single uhd device address args")
    parser.add_argument("-d", "--duration", default=10.0, type=float,
                        help="duration for the test in seconds")
    parser.add_argument("--rx_subdev", type=str, help="specify the device subdev for RX")
    parser.add_argument("--tx_subdev", type=str, help="specify the device subdev for TX")
    parser.add_argument("--rx_rate", type=float, help="specify to perform a RX rate test (sps)")
    parser.add_argument("--tx_rate", type=float, help="specify to perform a TX rate test (sps)")
    parser.add_argument("--rx_otw", type=str, default="sc16",
                        help="specify the over-the-wire sample mode for RX")
    parser.add_argument("--tx_otw", type=str, default="sc16",
                        help="specify the over-the-wire sample mode for TX")
    parser.add_argument("--rx_cpu", type=str, default="fc32",
                        help="specify the host/cpu sample mode for RX")
    parser.add_argument("--tx_cpu", type=str, default="fc32",
                        help="specify the host/cpu sample mode for TX")
    parser.add_argument("--rx_stream_args",
                        help="stream args for RX streamer", default="")
    parser.add_argument("--tx_stream_args", help="stream args for TX streamer",
                        default="")
    parser.add_argument("--ref", type=str,
                        help="clock reference (internal, external, mimo, gpsdo)")
    parser.add_argument("--pps", type=str, help="PPS source (internal, external, mimo, gpsdo)")
    parser.add_argument("--random", action="store_true", default=False,
                        help="Run with random values of samples in send() and recv() to stress-test"
                             " the I/O.")
    parser.add_argument("-c", "--channels", default=[0], nargs="+", type=int,
                        help="which channel(s) to use (specify \"0\", \"1\", \"0 1\", etc)")
    parser.add_argument("--rx_channels", nargs="+", type=int,
                        help="which RX channel(s) to use (specify \"0\", \"1\", \"0 1\", etc)")
    parser.add_argument("--tx_channels", nargs="+", type=int,
                        help="which TX channel(s) to use (specify \"0\", \"1\", \"0 1\", etc)")
    return parser.parse_args()


class LogFormatter(logging.Formatter):
    """Log formatter which prints the timestamp with fractional seconds"""
    @staticmethod
    def pp_now():
        """Returns a formatted string containing the time of day"""
        now = datetime.now()
        return "{:%H:%M}:{:05.2f}".format(now, now.second + now.microsecond / 1e6)
        # return "{:%H:%M:%S}".format(now)

    def formatTime(self, record, datefmt=None):
        converter = self.converter(record.created)
        if datefmt:
            formatted_date = converter.strftime(datefmt)
        else:
            formatted_date = LogFormatter.pp_now()
        return formatted_date


def setup_ref(usrp, ref, num_mboards):
    """Setup the reference clock"""
    if ref == "mimo":
        if num_mboards != 2:
            logger.error("ref = \"mimo\" implies 2 motherboards; "
                         "your system has %d boards", num_mboards)
            return False
        usrp.set_clock_source("mimo", 1)
    else:
        usrp.set_clock_source(ref)

    # Lock onto clock signals for all mboards
    if ref != "internal":
        logger.debug("Now confirming lock on clock signals...")
        end_time = datetime.now() + timedelta(milliseconds=CLOCK_TIMEOUT)
        for i in range(num_mboards):
            if ref == "mimo" and i == 0:
                continue
            is_locked = usrp.get_mboard_sensor("ref_locked", i)
            while (not is_locked) and (datetime.now() < end_time):
                time.sleep(1e-3)
                is_locked = usrp.get_mboard_sensor("ref_locked", i)
            if not is_locked:
                logger.error("Unable to confirm clock signal locked on board %d", i)
                return False
    return True


def setup_pps(usrp, pps, num_mboards):
    """Setup the PPS source"""
    if pps == "mimo":
        if num_mboards != 2:
            logger.error("ref = \"mimo\" implies 2 motherboards; "
                         "your system has %d boards", num_mboards)
            return False
        # make mboard 1 a slave over the MIMO Cable
        usrp.set_time_source("mimo", 1)
    else:
        usrp.set_time_source(pps)
    return True


def check_channels(usrp, args):
    """Check that the device has sufficient RX and TX channels available"""
    # Check RX channels
    if args.rx_rate:
        if args.rx_channels:
            rx_channels = args.rx_channels
        else:
            rx_channels = args.channels
        # Check that each channel specified is less than the number of total number of rx channels
        # the device can support
        dev_rx_channels = usrp.get_rx_num_channels()
        if not all(map((lambda chan: chan < dev_rx_channels), rx_channels)):
            logger.error("Invalid RX channel(s) specified.")
            return [], []
    else:
        rx_channels = []
    # Check TX channels
    if args.tx_rate:
        if args.tx_channels:
            tx_channels = args.tx_channels
        else:
            tx_channels = args.channels
        # Check that each channel specified is less than the number of total number of tx channels
        # the device can support
        dev_tx_channels = usrp.get_tx_num_channels()
        if not all(map((lambda chan: chan < dev_tx_channels), tx_channels)):
            logger.error("Invalid TX channel(s) specified.")
            return [], []
    else:
        tx_channels = []
    return rx_channels, tx_channels


def benchmark_rx_rate(usrp, rx_streamer, random, timer_elapsed_event, rx_statistics):
    """Benchmark the receive chain"""
    logger.info("Testing receive rate {:.3f} Msps on {:d} channels".format(
        usrp.get_rx_rate()/1e6, rx_streamer.get_num_channels()))

    # Make a receive buffer
    num_channels = rx_streamer.get_num_channels()
    max_samps_per_packet = rx_streamer.get_max_num_samps()
    # TODO: The C++ code uses rx_cpu type here. Do we want to use that to set dtype?
    recv_buffer = np.empty((num_channels, max_samps_per_packet), dtype=np.complex64)
    metadata = uhd.types.RXMetadata()

    # Craft and send the Stream Command
    stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.start_cont)
    stream_cmd.stream_now = (num_channels == 1)
    stream_cmd.time_spec = uhd.types.TimeSpec(usrp.get_time_now().get_real_secs() + INIT_DELAY)
    rx_streamer.issue_stream_cmd(stream_cmd)

    # To estimate the number of dropped samples in an overflow situation, we need the following
    # On the first overflow, set had_an_overflow and record the time
    # On the next ERROR_CODE_NONE, calculate how long its been since the recorded time, and use the
    #   tick rate to estimate the number of dropped samples. Also, reset the tracking variables
    had_an_overflow = False
    last_overflow = uhd.types.TimeSpec(0)
    # Setup the statistic counters
    num_rx_samps = 0
    num_rx_dropped = 0
    num_rx_overruns = 0
    num_rx_seqerr = 0
    num_rx_timeouts = 0
    num_rx_late = 0

    rate = usrp.get_rx_rate()
    # Receive until we get the signal to stop
    while not timer_elapsed_event.is_set():
        if random:
            stream_cmd.num_samps = np.random.randint(1, max_samps_per_packet+1, dtype=int)
            rx_streamer.issue_stream_cmd(stream_cmd)
        try:
            num_rx_samps += rx_streamer.recv(recv_buffer, metadata) * num_channels
        except RuntimeError as ex:
            logger.error("Runtime error in receive: %s", ex)
            return

        # Handle the error codes
        if metadata.error_code == uhd.types.RXMetadataErrorCode.none:
            # Reset the overflow flag
            if had_an_overflow:
                had_an_overflow = False
                num_rx_dropped += (metadata.time_spec - last_overflow).to_ticks(rate)
        elif metadata.error_code == uhd.types.RXMetadataErrorCode.overflow:
            had_an_overflow = True
            # Need to make sure that last_overflow is a new TimeSpec object, not
            # a reference to metadata.time_spec, or it would not be useful
            # further up.
            last_overflow = uhd.types.TimeSpec(
                metadata.time_spec.get_full_secs(),
                metadata.time_spec.get_frac_secs())
            # If we had a sequence error, record it
            if metadata.out_of_sequence:
                num_rx_seqerr += 1
            # Otherwise just count the overrun
            else:
                num_rx_overruns += 1
        elif metadata.error_code == uhd.types.RXMetadataErrorCode.late:
            logger.warning("Receiver error: %s, restarting streaming...", metadata.strerror())
            num_rx_late += 1
            # Radio core will be in the idle state. Issue stream command to restart streaming.
            stream_cmd.time_spec = uhd.types.TimeSpec(
                usrp.get_time_now().get_real_secs() + INIT_DELAY)
            stream_cmd.stream_now = (num_channels == 1)
            rx_streamer.issue_stream_cmd(stream_cmd)
        elif metadata.error_code == uhd.types.RXMetadataErrorCode.timeout:
            logger.warning("Receiver error: %s, continuing...", metadata.strerror())
            num_rx_timeouts += 1
        else:
            logger.error("Receiver error: %s", metadata.strerror())
            logger.error("Unexpected error on receive, continuing...")

    # Return the statistics to the main thread
    rx_statistics["num_rx_samps"] = num_rx_samps
    rx_statistics["num_rx_dropped"] = num_rx_dropped
    rx_statistics["num_rx_overruns"] = num_rx_overruns
    rx_statistics["num_rx_seqerr"] = num_rx_seqerr
    rx_statistics["num_rx_timeouts"] = num_rx_timeouts
    rx_statistics["num_rx_late"] = num_rx_late
    # After we get the signal to stop, issue a stop command
    rx_streamer.issue_stream_cmd(uhd.types.StreamCMD(uhd.types.StreamMode.stop_cont))


def benchmark_tx_rate(usrp, tx_streamer, random, timer_elapsed_event, tx_statistics):
    """Benchmark the transmit chain"""
    logger.info("Testing transmit rate %.3f Msps on %d channels",
                 usrp.get_tx_rate() / 1e6, tx_streamer.get_num_channels())

    # Make a transmit buffer
    num_channels = tx_streamer.get_num_channels()
    max_samps_per_packet = tx_streamer.get_max_num_samps()
    # TODO: The C++ code uses rx_cpu type here. Do we want to use that to set dtype?
    transmit_buffer = np.zeros((num_channels, max_samps_per_packet), dtype=np.complex64)
    metadata = uhd.types.TXMetadata()
    metadata.time_spec = uhd.types.TimeSpec(usrp.get_time_now().get_real_secs() + INIT_DELAY)
    metadata.has_time_spec = bool(num_channels)

    # Setup the statistic counters
    num_tx_samps = 0
    # TODO: The C++ has a single randomly sized packet sent here, then the thread returns
    num_timeouts_tx = 0
    # Transmit until we get the signal to stop
    if random:
        while not timer_elapsed_event.is_set():
            total_num_samps = np.random.randint(1, max_samps_per_packet + 1, dtype=int)
            num_acc_samps = 0
            while num_acc_samps < total_num_samps:
                num_tx_samps += tx_streamer.send(
                    transmit_buffer, metadata) * num_channels
                num_acc_samps += min(total_num_samps - num_acc_samps,
                                     tx_streamer.get_max_num_samps())
    else:
        while not timer_elapsed_event.is_set():
            try:
                num_tx_samps_now = tx_streamer.send(transmit_buffer, metadata) * num_channels
                num_tx_samps += num_tx_samps_now
                if num_tx_samps_now == 0:
                    num_timeouts_tx += 1
                    if (num_timeouts_tx % 10000) == 1:
                        logger.warning("Tx timeouts: %d", num_timeouts_tx)
                metadata.has_time_spec = False

            except RuntimeError as ex:
                logger.error("Runtime error in transmit: %s", ex)
                return

    tx_statistics["num_tx_samps"] = num_tx_samps

    # Send a mini EOB packet
    metadata.end_of_burst = True
    tx_streamer.send(np.zeros((num_channels, 0), dtype=np.complex64), metadata)


def benchmark_tx_rate_async_helper(tx_streamer, timer_elapsed_event, tx_async_statistics):
    """Receive and process the asynchronous TX messages"""
    async_metadata = uhd.types.TXAsyncMetadata()

    # Setup the statistic counters
    num_tx_seqerr = 0
    num_tx_underrun = 0
    num_tx_timeouts = 0  # TODO: Not populated yet
    try:
        while not timer_elapsed_event.is_set():
            # Receive the async metadata
            if not tx_streamer.recv_async_msg(async_metadata, 0.1):
                continue

            # Handle the error codes
            if async_metadata.event_code == uhd.types.TXMetadataEventCode.burst_ack:
                return
            if async_metadata.event_code in (
                    uhd.types.TXMetadataEventCode.underflow,
                    uhd.types.TXMetadataEventCode.underflow_in_packet):
                num_tx_underrun += 1
            elif async_metadata.event_code in (
                    uhd.types.TXMetadataEventCode.seq_error,
                    uhd.types.TXMetadataEventCode.seq_error_in_packet):
                num_tx_seqerr += 1
            else:
                logger.warning("Unexpected event on async recv (%s), continuing.",
                               async_metadata.event_code)
    finally:
        # Write the statistics back
        tx_async_statistics["num_tx_seqerr"] = num_tx_seqerr
        tx_async_statistics["num_tx_underrun"] = num_tx_underrun
        tx_async_statistics["num_tx_timeouts"] = num_tx_timeouts


def print_statistics(rx_statistics, tx_statistics, tx_async_statistics):
    """Print TRX statistics in a formatted block"""
    logger.debug("RX Statistics Dictionary: %s", rx_statistics)
    logger.debug("TX Statistics Dictionary: %s", tx_statistics)
    logger.debug("TX Async Statistics Dictionary: %s", tx_async_statistics)
    # Print the statistics
    statistics_msg = """Benchmark rate summary:
    Num received samples:     {}
    Num dropped samples:      {}
    Num overruns detected:    {}
    Num transmitted samples:  {}
    Num sequence errors (Tx): {}
    Num sequence errors (Rx): {}
    Num underruns detected:   {}
    Num late commands:        {}
    Num timeouts (Tx):        {}
    Num timeouts (Rx):        {}""".format(
        rx_statistics.get("num_rx_samps", 0),
        rx_statistics.get("num_rx_dropped", 0),
        rx_statistics.get("num_rx_overruns", 0),
        tx_statistics.get("num_tx_samps", 0),
        tx_async_statistics.get("num_tx_seqerr", 0),
        rx_statistics.get("num_rx_seqerr", 0),
        tx_async_statistics.get("num_tx_underrun", 0),
        rx_statistics.get("num_rx_late", 0),
        tx_async_statistics.get("num_tx_timeouts", 0),
        rx_statistics.get("num_rx_timeouts", 0))
    logger.info(statistics_msg)


def main():
    """Run the benchmarking tool"""
    args = parse_args()
    # Setup some argument parsing
    if not (args.rx_rate or args.tx_rate):
        logger.error("Please specify --rx_rate and/or --tx_rate")
        return False

    # Setup a usrp device
    usrp = uhd.usrp.MultiUSRP(args.args)
    if usrp.get_mboard_name() == "USRP1":
        logger.warning(
            "Benchmark results will be inaccurate on USRP1 due to insufficient features.")

    # Always select the subdevice first, the channel mapping affects the other settings
    if args.rx_subdev:
        usrp.set_rx_subdev_spec(uhd.usrp.SubdevSpec(args.rx_subdev))
    if args.tx_subdev:
        usrp.set_tx_subdev_spec(uhd.usrp.SubdevSpec(args.tx_subdev))

    logger.info("Using Device: %s", usrp.get_pp_string())

    # Set the reference clock
    if args.ref and not setup_ref(usrp, args.ref, usrp.get_num_mboards()):
        # If we wanted to set a reference clock and it failed, return
        return False

    # Set the PPS source
    if args.pps and not setup_pps(usrp, args.pps, usrp.get_num_mboards()):
        # If we wanted to set a PPS source and it failed, return
        return False
    # At this point, we can assume our device has valid and locked clock and PPS

    rx_channels, tx_channels = check_channels(usrp, args)
    if not rx_channels and not tx_channels:
        # If the check returned two empty channel lists, that means something went wrong
        return False
    logger.info("Selected %s RX channels and %s TX channels",
                rx_channels if rx_channels else "no",
                tx_channels if tx_channels else "no")

    logger.info("Setting device timestamp to 0...")
    # If any of these conditions are met, we need to synchronize the channels
    if args.pps == "mimo" or args.ref == "mimo" or len(rx_channels) > 1 or len(tx_channels) > 1:
        usrp.set_time_unknown_pps(uhd.types.TimeSpec(0.0))
    else:
        usrp.set_time_now(uhd.types.TimeSpec(0.0))

    threads = []
    # Make a signal for the threads to stop running
    quit_event = threading.Event()
    # Create a dictionary for the RX statistics
    # Note: we're going to use this without locks, so don't access it from the main thread until
    #       the worker has joined
    rx_statistics = {}
    # Spawn the receive test thread
    if args.rx_rate:
        usrp.set_rx_rate(args.rx_rate)
        st_args = uhd.usrp.StreamArgs(args.rx_cpu, args.rx_otw)
        st_args.channels = rx_channels
        st_args.args = uhd.types.DeviceAddr(args.rx_stream_args)
        rx_streamer = usrp.get_rx_stream(st_args)
        rx_thread = threading.Thread(target=benchmark_rx_rate,
                                     args=(usrp, rx_streamer, args.random, quit_event,
                                           rx_statistics))
        threads.append(rx_thread)
        rx_thread.start()
        rx_thread.setName("bmark_rx_stream")

    # Create a dictionary for the RX statistics
    # Note: we're going to use this without locks, so don't access it from the main thread until
    #       the worker has joined
    tx_statistics = {}
    tx_async_statistics = {}
    # Spawn the transmit test thread
    if args.tx_rate:
        usrp.set_tx_rate(args.tx_rate)
        st_args = uhd.usrp.StreamArgs(args.tx_cpu, args.tx_otw)
        st_args.channels = tx_channels
        st_args.args = uhd.types.DeviceAddr(args.tx_stream_args)
        tx_streamer = usrp.get_tx_stream(st_args)
        tx_thread = threading.Thread(target=benchmark_tx_rate,
                                     args=(usrp, tx_streamer, args.random, quit_event,
                                           tx_statistics))
        threads.append(tx_thread)
        tx_thread.start()
        tx_thread.setName("bmark_tx_stream")

        tx_async_thread = threading.Thread(target=benchmark_tx_rate_async_helper,
                                           args=(tx_streamer, quit_event, tx_async_statistics))
        threads.append(tx_async_thread)
        tx_async_thread.start()
        tx_async_thread.setName("bmark_tx_helper")

    # Sleep for the required duration
    # If we have a multichannel test, add some time for initialization
    if len(rx_channels) > 1 or len(tx_channels) > 1:
        args.duration += INIT_DELAY
    time.sleep(args.duration)
    # Interrupt and join the threads
    logger.debug("Sending signal to stop!")
    quit_event.set()
    for thr in threads:
        thr.join()

    print_statistics(rx_statistics, tx_statistics, tx_async_statistics)

    return True


if __name__ == "__main__":
    # Setup the logger with our custom timestamp formatting
    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    console = logging.StreamHandler()
    logger.addHandler(console)
    formatter = LogFormatter(fmt='[%(asctime)s] [%(levelname)s] (%(threadName)-10s) %(message)s')
    console.setFormatter(formatter)

    # Vamos, vamos, vamos!
    sys.exit(not main())



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
from . import libpyuhd as lib

ChdrPacket = lib.chdr.ChdrPacket
Endianness = lib.chdr.Endianness
ChdrWidth = lib.chdr.ChdrWidth
ChdrHeader = lib.chdr.ChdrHeader
CtrlPayload = lib.chdr.CtrlPayload
CtrlStatus = lib.chdr.CtrlStatus
PacketType = lib.chdr.PacketType
CtrlOpCode = lib.chdr.CtrlOpCode
MgmtPayload = lib.chdr.MgmtPayload
MgmtHop = lib.chdr.MgmtHop
MgmtOp = lib.chdr.MgmtOp
MgmtOpCode = lib.chdr.MgmtOpCode
MgmtOpSelDest = lib.chdr.MgmtOpSelDest
MgmtOpCfg = lib.chdr.MgmtOpCfg
MgmtOpNodeInfo = lib.chdr.MgmtOpNodeInfo
StrsPayload = lib.chdr.StrsPayload
StrsStatus = lib.chdr.StrsStatus
StrcPayload = lib.chdr.StrcPayload
StrcOpCode = lib.chdr.StrcOpCode

def __get_payload(self):
    pkt_type = self.get_header().pkt_type
    if pkt_type == PacketType.MGMT:
        return self.get_payload_mgmt()
    elif pkt_type == PacketType.CTRL:
        return self.get_payload_ctrl()
    elif pkt_type == PacketType.STRC:
        return self.get_payload_strc()
    elif pkt_type == PacketType.STRS:
        return self.get_payload_strs()
    elif pkt_type in {PacketType.DATA_NO_TS, PacketType.DATA_WITH_TS}:
        raise RuntimeError("Cannot deserialize a Data Payload")
    else:
        raise RuntimeError("Invalid pkt_type in ChdrHeader")

ChdrPacket.get_payload = __get_payload

def __to_string_with_payload(self):
    pkt_type = self.get_header().pkt_type
    if pkt_type == PacketType.MGMT:
        return self.to_string_with_payload_mgmt()
    elif pkt_type == PacketType.CTRL:
        return self.to_string_with_payload_ctrl()
    elif pkt_type == PacketType.STRC:
        return self.to_string_with_payload_strc()
    elif pkt_type == PacketType.STRS:
        return self.to_string_with_payload_strs()
    elif pkt_type in {PacketType.DATA_NO_TS, PacketType.DATA_WITH_TS}:
        return self.__str__()
    else:
        raise RuntimeError("Invalid pkt_type in ChdrHeader")

ChdrPacket.to_string_with_payload = __to_string_with_payload
